"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.downloadJatsFromUrl = exports.checkIfPubMedCentralHasJats = exports.convertPMID2PMCID = void 0;
const doi_utils_1 = __importDefault(require("doi-utils"));
const chalk_1 = __importDefault(require("chalk"));
const node_fetch_1 = __importDefault(require("node-fetch"));
const myst_cli_utils_1 = require("myst-cli-utils");
const fair_principles_1 = require("fair-principles");
const resolvers_1 = require("./resolvers");
function logAboutJatsFailing(session, jatsUrls) {
    session.log.warn('⛔️ JATS may not be Open Access 😭, you should petition your local representative 🪧');
    session.log.info(`${chalk_1.default.green(`\nThe XML ${chalk_1.default.bold('should')} be here:\n\n${jatsUrls.join('\n')}`)}\n`);
    const FAIR = (0, fair_principles_1.highlightFAIR)('A', { chalk: chalk_1.default });
    session.log.info(`Some publishers aggressively block programmatic access, which isn't ${FAIR}.`);
    session.log.debug((0, fair_principles_1.formatPrinciples)('A*', { chalk: chalk_1.default }));
    session.log.info(`${chalk_1.default.blue('The link may work in a browser.')}\n`);
}
function dowloadFromUrl(session, jatsUrl) {
    return __awaiter(this, void 0, void 0, function* () {
        const toc = (0, myst_cli_utils_1.tic)();
        session.log.debug(`Fetching JATS from ${jatsUrl}`);
        const resp = yield (0, node_fetch_1.default)(jatsUrl, {
            headers: [
                ['accept', 'application/xml'],
                [
                    'user-agent',
                    // A bunch of publishers just show the login screen or quickly block you.
                    // We don't want to DDOS these publishers, they are the _good ones_ for sharing the XML!!
                    // But some block on the second request?!
                    // So we can pretend to be a random browser, I guess. How silly. 🤷‍♂️
                    `Mozilla/5.0 (Macintosh; Intel Mac OS X ${Math.floor(Math.random() * 100)})`,
                ],
            ],
        });
        if (!resp.ok) {
            session.log.debug(`JATS failed to download from "${jatsUrl}"`);
            throw new Error(`STATUS ${resp.status}: ${resp.statusText}`);
        }
        const contentType = resp.headers.get('content-type');
        if (!((contentType === null || contentType === void 0 ? void 0 : contentType.includes('application/xml')) ||
            (contentType === null || contentType === void 0 ? void 0 : contentType.includes('text/xml')) ||
            (contentType === null || contentType === void 0 ? void 0 : contentType.includes('text/plain')))) {
            session.log.warn(`Expected content-type "application/xml" instead we got "${contentType}" for ${jatsUrl}\n${chalk_1.default.dim('Things may not work, but we are going to try our best...')}`);
        }
        const data = yield resp.text();
        session.log.debug(toc(`Fetched document with content-type "${contentType}" in %s`));
        return data;
    });
}
/**
 * There are 5.8M or so DOIs that have a full XML record:
 *
 * https://api.crossref.org/works?filter=full-text.type:application/xml,full-text.application:text-mining&facet=publisher-name:*&rows=0
 *
 * This function tries to find the correct URL for the record.
 */
function checkIfDoiHasJats(session, urlOrDoi) {
    var _a, _b, _c, _d;
    return __awaiter(this, void 0, void 0, function* () {
        if (!doi_utils_1.default.validate(urlOrDoi))
            return;
        const toc = (0, myst_cli_utils_1.tic)();
        const doiUrl = doi_utils_1.default.buildUrl(urlOrDoi);
        session.log.debug(`Attempting to resolving full XML from DOI ${doiUrl}`);
        const resp = yield (0, node_fetch_1.default)(doiUrl, { headers: [['Accept', 'application/json']] });
        if (!resp.ok) {
            // Silently return -- other functions can try!
            session.log.debug(`DOI failed to resolve: ${doiUrl}`);
            return;
        }
        const data = (yield resp.json());
        session.log.debug(toc(`DOI resolved in %s with ${(_b = (_a = data.link) === null || _a === void 0 ? void 0 : _a.length) !== null && _b !== void 0 ? _b : 0} links to content`));
        if (data.link) {
            session.log.debug(['', ...data.link.map((link) => `content-type: ${link['content-type']}, ${link.URL}\n`)].join('  - '));
        }
        const fullXml = (_d = (_c = data.link) === null || _c === void 0 ? void 0 : _c.find((link) => { var _a; return ['text/xml', 'application/xml'].includes((_a = link['content-type']) !== null && _a !== void 0 ? _a : ''); })) === null || _d === void 0 ? void 0 : _d.URL;
        if (fullXml)
            return fullXml;
        session.log.debug(`Could not find XML in DOI record ${doiUrl}`);
        return undefined;
    });
}
/**
 * https://www.ncbi.nlm.nih.gov/pmc/tools/id-converter-api/
 */
function convertPMID2PMCID(session, PMID) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        if (PMID.startsWith('https://')) {
            const idPart = new URL(PMID).pathname.slice(1);
            session.log.debug(`Extract ${PMID} to ${idPart}`);
            return convertPMID2PMCID(session, idPart);
        }
        const toc = (0, myst_cli_utils_1.tic)();
        const converter = 'https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/';
        const resp = yield (0, node_fetch_1.default)(`${converter}?tool=jats-xml&format=json&ids=${PMID}`, {
            headers: [['Accept', 'application/json']],
        });
        if (!resp.ok) {
            // Silently return -- other functions can try!
            session.log.debug(`Failed to convert PubMedID: ${PMID}`);
            return;
        }
        const data = yield resp.json();
        const PMCID = (_b = (_a = data === null || data === void 0 ? void 0 : data.records) === null || _a === void 0 ? void 0 : _a[0]) === null || _b === void 0 ? void 0 : _b.pmcid;
        session.log.debug(toc(`Used nih.gov to transform ${PMID} to ${PMCID} in %s.`));
        return PMCID;
    });
}
exports.convertPMID2PMCID = convertPMID2PMCID;
function pubMedCentralJats(PMCID) {
    const normalized = PMCID.replace(/^PMC:?/, '');
    return `https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pmc&id=${normalized}`;
}
function checkIfPubMedCentralHasJats(session, urlOrDoi) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        if (urlOrDoi.match(/^PMC:?([0-9]+)$/))
            return pubMedCentralJats(urlOrDoi);
        if (!doi_utils_1.default.validate(urlOrDoi))
            return;
        const toc = (0, myst_cli_utils_1.tic)();
        const doiUrl = doi_utils_1.default.buildUrl(urlOrDoi);
        session.log.debug(`Attempting to resolve PMCID using OpenAlex from ${doiUrl}`);
        const openAlexUrl = `https://api.openalex.org/works/${doiUrl}`;
        const resp = yield (0, node_fetch_1.default)(openAlexUrl, { headers: [['Accept', 'application/json']] });
        if (!resp.ok) {
            // Silently return -- other functions can try!
            session.log.debug(`Failed to lookup on OpenAlex: ${openAlexUrl}`);
            return;
        }
        const data = (yield resp.json());
        const PMID = (_a = data === null || data === void 0 ? void 0 : data.ids) === null || _a === void 0 ? void 0 : _a.pmid;
        let PMCID = (_b = data === null || data === void 0 ? void 0 : data.ids) === null || _b === void 0 ? void 0 : _b.pmcid;
        if (!PMCID && !!PMID) {
            session.log.debug(toc(`OpenAlex resolved ${data === null || data === void 0 ? void 0 : data.ids.openalex} in %s. There is no PMCID, but there is a PMID`));
            PMCID = yield convertPMID2PMCID(session, PMID);
            if (!PMCID) {
                session.log.debug(toc(`PubMed does not have a record of ${PMID}`));
                return;
            }
        }
        if (!PMCID) {
            session.log.debug(toc(`OpenAlex resolved ${data === null || data === void 0 ? void 0 : data.ids.openalex} in %s, but there is no PMCID`));
            return;
        }
        session.log.debug(toc(`OpenAlex resolved in %s, with a PMCID of ${PMCID}`));
        return pubMedCentralJats(PMCID);
    });
}
exports.checkIfPubMedCentralHasJats = checkIfPubMedCentralHasJats;
function downloadJatsFromUrl(session, urlOrDoi, resolvers) {
    return __awaiter(this, void 0, void 0, function* () {
        const expectedUrls = (yield Promise.all([
            checkIfPubMedCentralHasJats(session, urlOrDoi),
            checkIfDoiHasJats(session, urlOrDoi),
        ])).filter((u) => !!u);
        if (expectedUrls.length > 0) {
            session.log.debug(['Trying URLs:\n', ...expectedUrls.map((url) => ` ${url}\n`)].join('  - '));
            for (let index = 0; index < expectedUrls.length; index++) {
                const url = expectedUrls[index];
                try {
                    const data = yield dowloadFromUrl(session, url);
                    if (data)
                        return { source: url, data };
                }
                catch (error) {
                    session.log.debug(error.message);
                }
            }
            // If there are expected URLs that don't work: see something, say something, etc.
            logAboutJatsFailing(session, expectedUrls);
        }
        if (doi_utils_1.default.validate(urlOrDoi)) {
            const jatsUrl = yield (0, resolvers_1.customResolveJatsUrlFromDoi)(session, urlOrDoi, resolvers);
            const data = yield dowloadFromUrl(session, jatsUrl);
            return { source: jatsUrl, data };
        }
        if ((0, myst_cli_utils_1.isUrl)(urlOrDoi)) {
            session.log.debug("No resolver matched, and the URL doesn't look like a DOI. We will attempt to download it directly.");
            const data = yield dowloadFromUrl(session, urlOrDoi);
            return { source: urlOrDoi, data };
        }
        throw new Error(`Could not find ${urlOrDoi} locally, and it doesn't look like a URL or DOI`);
    });
}
exports.downloadJatsFromUrl = downloadJatsFromUrl;
